/*----------------------------------------------------------------------
  File    : npi.c
  Contents: naive possibilistic classifier induction
  Author  : Christian Borgelt
  History : 2001.02.07 file created from file nbi.c
            2001.07.17 adapted to modified module scan
            2003.08.16 slight changes in error message output
            2007.02.14 adapted to modified module attset
            2007.10.10 evaluation of attribute directions added
            2011.12.15 processing without table reading improved
            2013.08.09 adapted to definitions ATTID, VALID, TPLID etc.
            2013.08.29 adapted to new function as_target()
            2013.12.16 bug concerning metric attributes in input fixed
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <assert.h>
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_DESC
#define AS_DESC
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#include "table.h"
#ifndef NPC_INDUCE
#define NPC_INDUCE
#endif
#include "nposs.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "npi"
#define DESCRIPTION "naive possibilistic classifier induction"
#define VERSION     "version 2.1 (2014.10.24)         " \
                    "(c) 1999-2014   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION    (-6)        /* unknown option */
#define E_OPTARG    (-7)        /* missing option argument */
#define E_ARGCNT    (-8)        /* wrong number of arguments */
#define E_PARSE     (-9)        /* parse error on domain file */
#define E_ATTCNT   (-10)        /* no usable attributes found */
#define E_UNKCLS   (-11)        /* unknown class attribute */
#define E_MULCLS   (-12)        /* multiple class attributes */
#define E_CLSTYPE  (-13)        /* class attribute is not nominal */
#define E_SIMP     (-14)        /* unknown simplification mode */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_ATTCNT  -10 */  "no (usable) attributes (need at least 1)",
  /* E_UNKCLS  -11 */  "unknown class attribute '%s'",
  /* E_MULCLS  -12 */  "multiple class attributes found",
  /* E_CLSTYPE -13 */  "class attribute '%s' is not nominal",
  /* E_BALANCE -14 */  "unknown balancing mode %c",
  /* E_SIMP    -15 */  "unknown simplification mode %c",
  /*           -16 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static SCANNER *scan   = NULL;  /* scanner */
static TABREAD *tread  = NULL;  /* table reader */
static ATTSET  *attset = NULL;  /* attribute set */
static TABLE   *table  = NULL;  /* data table */
static NPC     *npc    = NULL;  /* naive possibilistic classifier */
static FILE    *out    = NULL;  /* classifier output file */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (npc)    npc_delete(npc, 0);   \
  if (attset) as_delete(attset);    \
  if (table)  tab_delete(table, 0); \
  if (scan)   scn_delete(scan,  1); \
  if (out && (out != stdout)) fclose(out);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counter */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_dom  = NULL;      /* name of domain     file */
  CCHAR   *fn_hdr  = NULL;      /* name of header     file */
  CCHAR   *fn_tab  = NULL;      /* name of table      file */
  CCHAR   *fn_npc  = NULL;      /* name of classifier file */
  CCHAR   *recseps = NULL;      /* record     separators */
  CCHAR   *fldseps = NULL;      /* field      separators */
  CCHAR   *blanks  = NULL;      /* blank      characters */
  CCHAR   *nullchs = NULL;      /* null value characters */
  CCHAR   *comment = NULL;      /* comment    characters */
  CCHAR   *clsname = NULL;      /* name of the class attribute */
  ATTID   clsid    = -1;        /* id of class attribute */
  int     mode     = AS_ATT|AS_NOXATT; /* table file read mode */
  int     simp     = 0;         /* flag for classifier simplification */
  int     setup    = 0;         /* setup/induction mode */
  int     desc     = 0;         /* description mode */
  int     maxlen   = 0;         /* maximal output line length */
  ATT     *att;                 /* to traverse the attributes */
  ATTID   m;                    /* number of attributes */
  TPLID   n;                    /* number of data tuples */
  double  w;                    /* weight of data tuples */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument given */
    printf("usage: %s [options] domfile "
                     "[-d|-h hdrfile] tabfile npcfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-c#      name of class attribute/field          "
                    "(default: last attribute)\n");
    printf("-s#      simplify classifier                    "
                    "(default: all attributes)\n"
           "         (a: by adding attributes, "
                     "r: by removing attributes)\n");
    printf("-p       print degrees of possibility (percent) "
                    "(default: only absolute)\n");
    printf("-l#      output line length                     "
                    "(default: no limit)\n");
    printf("-r#      record     separators                  "
                    "(default: \"\\n\")\n");
    printf("-f#      field      separators                  "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank      characters                  "
                    "(default: \" \\t\\r\")\n");
    printf("-u#      null value characters                  "
                    "(default: \"?*\")\n");
    printf("-C#      comment    characters                  "
                    "(default: \"#\")\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("domfile  file containing domain descriptions\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header  "
                    "(field names) from hdrfile\n");
    printf("hdrfile  file containing table header "
                    "(field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("npcfile  file to write naive possibilistic "
                    "classifier to\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse the arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'c': optarg  = &clsname;              break;
          case 's': simp    = (*s) ? *s++ : 0;       break;
          case 'p': desc   |= NPC_POSS;              break;
          case 'l': maxlen  = (int)strtol(s, &s, 0); break;
          case 'r': optarg  = &recseps;              break;
          case 'f': optarg  = &fldseps;              break;
          case 'b': optarg  = &blanks;               break;
          case 'C': optarg  = &comment;              break;
          case 'u': optarg  = &nullchs;              break;
          case 'n': mode   |= AS_WEIGHT;             break;
          case 'd': mode   |= AS_DFLT;               break;
          case 'h': optarg  = &fn_hdr;               break;
          default : error(E_OPTION, *--s);           break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_dom = s;      break;
        case  1: fn_tab = s;      break;
        case  2: fn_npc = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check the option argument */
  if (k != 3) error(E_ARGCNT);  /* and the number of arguments */
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" */
  i = ( fn_hdr && !*fn_hdr) ? 1 : 0;
  if  (!fn_dom || !*fn_dom) i++;
  if  (!fn_tab || !*fn_tab) i++;/* check assignments of stdin: */
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  if      (simp == 'a') setup |= NPC_ADD;
  else if (simp == 'r') setup |= NPC_REMOVE;
  else if (simp !=  0 )         /* check simplification mode */
    error(E_SIMP, simp);        /* (must be 'add' or 'remove') */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- parse domain descriptions --- */
  attset = as_create("domains", att_delete);
  if (!attset) error(E_NOMEM);  /* create an attribute set */
  scan = scn_create();          /* create a scanner */
  if (!scan)   error(E_NOMEM);  /* for the domain file */
  t = clock();                  /* start timer, open input file */
  if (scn_open(scan, NULL, fn_dom) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  if ((as_parse(attset, scan, AT_ALL, 1) != 0)
  ||  !scn_eof(scan, 1))        /* parse domain descriptions */
    error(E_PARSE, scn_name(scan));
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* clear the scanner variable */
  m    = as_attcnt(attset);     /* get the number of attributes */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  if (m <= 0) error(E_ATTCNT);  /* check for at least one attribute */

  /* --- determine id of class attribute --- */
  clsid = as_target(attset, clsname, 1);
  if (clsid < 0) error(clsname ? E_UNKCLS : E_MULCLS, clsname);
  if (att_type(as_att(attset, clsid)) != AT_NOM)
    error(E_CLSTYPE, att_name(as_att(attset, clsid)));
  for (i = as_attcnt(attset); --i >= 0; ) {
    att = as_att(attset, i);    /* mark all non-nominal attributes */
    att_setmark(att, (att_type(att) == AT_NOM) ? -1 : 0);
  }                             /* cut attributes and get class id. */
  as_attcut(NULL, attset, AS_MARKED);
  clsid = as_target(attset, clsname, 1);

  /* --- read table header --- */
  tread = trd_create();         /* create a table reader and */
  if (!tread) error(E_NOMEM);   /* configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, nullchs, comment);
  if (fn_hdr) {                 /* if a header file is given */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_hdr) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_close(tread);           /* read table header, close file */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    mode &= ~(AS_ATT|AS_DFLT);  /* print a success message and */
  }                             /* remove the attribute flag */

  /* --- read data table --- */
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_tab) != 0)
    error(E_FOPEN, trd_name(tread));
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  table = tab_create("table", attset, tpl_delete);
  if (!table) error(E_NOMEM);   /* create a data table */
  k = tab_read(table, tread, mode);
  if (k < 0) error(-k, tab_errmsg(table, NULL, 0));
  trd_delete(tread, 1);         /* read the table body and */
  tread = NULL;                 /* delete the table reader */
  m = tab_attcnt(table);        /* get the number of attributes */
  n = tab_tplcnt(table);        /* and the number of data tuples */
  w = tab_tplwgt(table);        /* and print a success message */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
  t = clock();                  /* start timer, print log message */
  fprintf(stderr, "reducing table ... ");
  n = tab_reduce(table);        /* reduce table for speed up */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

  /* --- build naive possibilistic classifier --- */
  t = clock();                  /* start the timer */
  fprintf(stderr, "building classifier ... ");
  npc = npc_induce(table, clsid, setup);
  if (!npc) error(E_NOMEM);     /* induce a possibilistic classifier */
  m = npc_mark(npc);            /* and mark the selected attributes */
  fprintf(stderr, "[%"ATTID_FMT"+1 attribute(s)]", m-1);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- describe created classifier --- */
  t = clock();                  /* start the timer */
  if (fn_npc && (strcmp(fn_npc, "-") == 0)) fn_npc = "";
  if (fn_npc && *fn_npc) { out = fopen(fn_npc, "w"); }
  else                   { out = stdout; fn_npc = "<stdout>"; }
  fprintf(stderr, "writing %s ... ", fn_npc);
  if (!out) error(E_FOPEN, fn_npc);
  k = (simp) ? AS_MARKED : 0;   /* print only class and marked atts. */
  if (as_desc(attset, out, k|AS_TITLE, maxlen) != 0)
    error(E_FWRITE, fn_npc);    /* describe attribute domains */
  fputc('\n', out);             /* leave one line empty */
  if (npc_desc(npc, out, desc|NPC_TITLE|NPC_INFO, maxlen) != 0)
    error(E_FWRITE, fn_npc);    /* describe possibilistic classifier */
  if (out && (((out == stdout) ? fflush(out) : fclose(out)) != 0))
    error(E_FWRITE, fn_npc);    /* close the output file and */
  out = NULL;                   /* print a success message */
  fprintf(stderr, "[%"ATTID_FMT"+1 attribute(s)]", m-1);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
